#include "fcgi_stdio.h"
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <syslog.h>
#include <semaphore.h>
#include <time.h>
#include <signal.h>
#include <pthread.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>
#include <unistd.h>
#include <dirent.h>
#include "libpq-fe.h"
#include "pg_report.h"
#include "reporting_config.h"
#include "inno-lib.h"
#include "cdr.h"
#include "mypbx.h"

#define LISTENSOCK_FILENO 0
#define LISTENSOCK_FLAGS 0

#define WEBDAV_LOG      "/webdav/log/"
#define WEBDAV_CDR      "/webdav/cdr/"
#define WEBDAV_ALARM    "/webdav/alarm/"
#define DOC_ROOT        "/var/www/innovaphone"

struct _cdr_node {
    struct _cdr_node *next;
    struct _cdr *cdr;
    int db_ok;
    FCGX_Request *fcgi_req;
    int ind;
} cdr_node_t;

typedef struct _req {
    FCGX_Request *fcgi_req;
    pthread_t tid;
    int thread_running;
    int ind;
    pthread_mutex_t *cdr_queue_mutex;
    sem_t *cdr_queue_count;
    struct _cdr_node **cdr_queue;
    sem_t cdr_sem;
    PGconn **pg_conn;
    ldap_servers_t **ldap;
    pthread_mutex_t *ldap_mutex;
} req_t;

typedef struct _psql_ {
    pthread_t tid;
    pthread_mutex_t *cdr_queue_mutex;
    sem_t *cdr_queue_count;
    struct _cdr_node **cdr_queue;
    PGconn *conn;
} psql_t;

static int logging(char *filename, char *query_string)
{
    char * event = NULL, * time = NULL, * date = NULL, * from = NULL, * msg = NULL; 
    if(!*query_string) {
        do_log(LOG_LOG, "", "logging: query string empty");
        return 0;
    }
    if(*query_string == '<') {
        msg = str_malloc(query_string);
    }
    else {
        CURL * conn = curl_easy_init();
        if(conn) {
            event = get_url_param(conn, query_string, "event", "logging", LOG_LOG);
            time = get_url_param(conn, query_string, "time", "logging", LOG_LOG);
            date = get_url_param(conn, query_string, "date", "logging", LOG_LOG);
            from = get_url_param(conn, query_string, "from", "logging", LOG_LOG);
            msg = get_url_param(conn, query_string, "msg", "logging", LOG_LOG);
            curl_easy_cleanup(conn);
        }
    }
    if(!msg) {
        if(msg) free(msg);
        if(date) free(date);
        if(from) free(from);
        if(event) free(event);
        if(time) free(time);
        return NO_MSG_FIELD_LOG_REQ;
    }
    if(!from) from = str_malloc("");
    if(!date || strlen(date) != 15) {
        if(date) free(date);
        date = str_malloc("---------------");
    }
    if(!time) time = str_malloc("----------");
    if(!event) event = str_malloc("syslog");

    do_log(LOG_LOG, "", "%.4s-%.2s-%.2s %.2s:%.2s:%.2s [ %s ] [ %s] [ %s ] %s",
        &date[0], 
        &date[4],
        &date[6],
        &date[9],
        &date[11],
        &date[13],
        time,
        event,
        from,
        msg);
    
    free(msg);
    free(date);
    free(from);
    free(event);
    free(time);

    return 0;
}

static const char * str_alarms = "alarms_xml_post";
static const char * str_fault = "fault_xml_post";

static int alarms(char *filename, char *query_string, char * method)
{
    // waiting for "fault_xml_post" or "alarms_xml_post"
    if(!strncmp(query_string, str_fault, strlen(str_fault))) {
        query_string += strlen(str_fault) + 1;
    }
    else if(!strncmp(query_string, str_alarms, strlen(str_alarms))) {
        query_string += strlen(str_alarms) + 1;
    }
    if(!strcmp(method, "GET")) {
        CURL * conn = curl_easy_init();
        if(conn) {
            char * msg = get_url_param(conn, query_string, NULL, "alarm", LOG_ALARMS);
            curl_easy_cleanup(conn);
            do_log(LOG_ALARMS, "", "%s", msg);
            free(msg);
        }
    }
    else {
        do_log(LOG_ALARMS, "", "%s", query_string);
    }
    return 0;
}

static int relay_cdr(char * filename, char *query_string)
{
    do_log(LOG_CDRS, "", "%s", query_string);
    return 0;
}

static void create_ldap_changed_file(void)
{
    FILE *fp;

    if((fp = fopen(LDAP_SERVER_FILE, "r")) == NULL) { // file does not exist, create it
        if((fp = fopen(LDAP_SERVER_FILE, "w")) != NULL) {
            fclose(fp);
        }
        else
            do_log(LOG_INNOVAPHONE, "", "error creating ldap changed file");
    }
    else
        fclose(fp);
}

static int cdr(req_t *req, char *query_string)
{
    int ret = 0;
    struct _cdr_node * cdr_node = (struct _cdr_node *)malloc(sizeof(struct _cdr_node));
    struct _cdr * cdr = (struct _cdr *)malloc(sizeof(struct _cdr));
    struct timeval tvBegin, tvEnd, tvDiff;
    struct _cdr_node *pTmp, *pTmp0;

    if(trace_level & TRACE_TIME) {
        gettimeofday(&tvEnd, NULL);
    }

    if(trace_level & TRACE_STD) {
        do_log(LOG_REPORTING, "", "[%d] process an incoming cdr", req->ind);
    }

    if(!cdr_node || !cdr) {
        do_log(LOG_REPORTING, "", "cdr: no memory available");
        return NO_MEMORY_FOR_POST_REQ;
    }

    cdr->ldap_mutex = req->ldap_mutex;
    cdr->ldap_act = req->ldap;
    cdr->sem = &req->cdr_sem;
    cdr->ev_prep_query[0] = PG_1EVENTS_PREP_TRANS;
    cdr->ev_prep_query[1] = PG_2EVENTS_PREP_TRANS;
    cdr->ev_prep_query[2] = PG_3EVENTS_PREP_TRANS;
    cdr->ev_prep_query[3] = PG_4EVENTS_PREP_TRANS;
    cdr->ev_prep_query[4] = PG_5EVENTS_PREP_TRANS;
    cdr->grp_prep_query[0] = PG_1GROUPS_PREP_TRANS;
    cdr->grp_prep_query[1] = PG_2GROUPS_PREP_TRANS;
    cdr->grp_prep_query[2] = PG_3GROUPS_PREP_TRANS;
    cdr->grp_prep_query[3] = PG_4GROUPS_PREP_TRANS;
    cdr->grp_prep_query[4] = PG_5GROUPS_PREP_TRANS;
    cdr_node->ind = req->ind;
    cdr_node->db_ok = 0;
    cdr_node->cdr = cdr;
    cdr_node->fcgi_req = req->fcgi_req;
    cdr_node->next = NULL;

    ret = parse_cdr(cdr, query_string, req->fcgi_req);
    if(ret) {
        free_cdr_data(cdr);
        free(cdr);
        free(cdr_node);

        if(ret == ERROR_UNLICENSED_CDR) return 0;

        return ret;
    }

    pthread_mutex_lock(req->cdr_queue_mutex);
    
    //do_log(LOG_REPORTING, "", "[%d] enter mutex semaphore %x cdrs=%d", req->ind, req->cdr_queue_mutex, cdr_node->cdr->num_cdrs);

    if(*req->cdr_queue == NULL) {
        *req->cdr_queue = cdr_node;
    }
    else {
        pTmp = *req->cdr_queue;
        while(pTmp != NULL) {
            pTmp0 = pTmp;
            pTmp = pTmp->next;
        }
        pTmp0->next = cdr_node;
    }

    sem_post(req->cdr_queue_count);

    //do_log(LOG_REPORTING, "", "[%d] leave mutex semaphore", req->ind);

    pthread_mutex_unlock(req->cdr_queue_mutex);

    sem_wait(&req->cdr_sem);

    ret = cdr_node->db_ok;
    if(trace_level & TRACE_STD) {
        if(ret) do_log(LOG_REPORTING, "", "Failed to insert CDR=%d", ret);
    }

    free_cdr_data(cdr);
    free(cdr);
    free(cdr_node);

    if(trace_level & TRACE_TIME) {
        gettimeofday(&tvEnd, NULL);
        timeval_subtract(&tvDiff, &tvEnd, &tvBegin);
        do_log(LOG_REPORTING, "", "cdr insert: %ld.%06ld", tvDiff.tv_sec, tvDiff.tv_usec);
    }

    return ret;
}

void postgresql_exec(void *a)
{
    psql_t *pg = (psql_t *)a;
    struct _cdr_node **q = pg->cdr_queue;
    int i;

    while(1) {
        struct _cdr_node* next_cdr;

        // waits if cdr_queue_count is zero
        sem_wait(pg->cdr_queue_count);

        pthread_mutex_lock(pg->cdr_queue_mutex);
        //do_log(LOG_REPORTING, "", "enter semaphore and mutex= %x %x", q, *q);
        next_cdr = *q;
        *q = next_cdr->next;
        pthread_mutex_unlock(pg->cdr_queue_mutex);

        for(i=0; i<next_cdr->cdr->num_cdrs; i++) {
            if(next_cdr->cdr->cdrs[i]) {
                ConnStatusType pg_st;
                int conn_db_error = 0;
                int ret_db;

                //do_log(LOG_REPORTING, "", "[%x] proc cdr = %x", next_cdr->ind, next_cdr);

                if(pg->conn == NULL)
                    pg_st = CONNECTION_BAD;
                else
                    pg_st = PQstatus(pg->conn);

                if(pg_st != CONNECTION_OK) {
                    do_log(LOG_REPORTING, "", "Connection to database lost %d", pg_st);
                    if(pg->conn) PQfinish(pg->conn);
                    if((conn_db_error = init_db_conn(&pg->conn, NULL)) == 0) {
                        pg_st = PQstatus(pg->conn);
                    }
                    if(pg->conn == NULL) create_ldap_changed_file();
                }

                if(pg_st == CONNECTION_OK) {
                    if((ret_db = send_cdr_to_db(next_cdr->cdr->cdrs[i], next_cdr->fcgi_req, pg->conn)) == ERROR_DB_HANDLER_NOT_VALID) {
                        do_log(LOG_REPORTING, "", "Restarting connection to database");
                        if((conn_db_error = init_db_conn(&pg->conn, NULL)) == 0) {
                            pg_st = PQstatus(pg->conn);
                        }
                        if(pg->conn == NULL) create_ldap_changed_file();
                    }
                    next_cdr->db_ok += ret_db;
                }
                else {
                    next_cdr->db_ok = conn_db_error;
                    break;
                }
            }
            else
                do_log(LOG_REPORTING, "", "[%d] No data available for cdr!", next_cdr->ind);
        }

        // tell cdr we are done processing 
        sem_post(next_cdr->cdr->sem);

        //do_log(LOG_REPORTING, "", "end semaphore %x %d", next_cdr->cdr->sem, *next_cdr->cdr->sem);
    }

    close_db_conn(&pg->conn);
    pthread_exit((void*)0);
}

void process_fcgi(void *a)
{
    req_t *req = (req_t *)a;
    FCGX_Request *cgi = (FCGX_Request *)req->fcgi_req;
    int content_length, proc_error = 0, bytes_sent = 0;
    char *req_method = NULL;
    char *query_string = NULL;
    char *conn_length = NULL;
    char *post = NULL;
    char *filename = NULL, *doc_root = NULL, *script_name = NULL;
    char *logfile = NULL;
    int log_type = LOG_INNOVAPHONE;
    int external_cmd = 0;
    int output_len = 0;
    script_name = FCGX_GetParam("SCRIPT_NAME", cgi->envp);
    doc_root = FCGX_GetParam("DOCUMENT_ROOT", cgi->envp);
    int reporting_installed = 0;
    DIR * reporting_dir = NULL;
    
    logfile = (char *)malloc(256*sizeof(char));
    if((doc_root == NULL) || (script_name == NULL) || (logfile == NULL)) {
        do_log(LOG_INNOVAPHONE, "", "process.fcgi: document root or script name is empty: %s %s", doc_root, script_name);
        proc_error = ERROR_DOC_ROOT_SCRIPT_NAME_EMPTY;
        if(logfile) free(logfile);
    }
    else {
        strcpy(logfile, doc_root);
        strcat(logfile, "/apps/innovaphone-reporting/");
        reporting_dir = opendir(logfile);
        if(reporting_dir) {
            closedir(reporting_dir);
            reporting_installed = 1;
        }
        if(reporting_installed && (strstr(script_name, "innovaphone-reporting") || strstr(script_name, "cdr"))) {
            log_type = LOG_REPORTING;
        }
        req_method = FCGX_GetParam("REQUEST_METHOD", cgi->envp);
        if(!strcmp(req_method, "GET")) {
            if(trace_level & TRACE_STD) {
                do_log(log_type, "", "process.fcgi: process GET request");
            }
            query_string = FCGX_GetParam("QUERY_STRING", cgi->envp);
            if(trace_level & TRACE_STD) {
                do_log(log_type, "", "GET[%d]: %s", strlen(query_string), query_string);
            }
        } else if(!strcmp(req_method, "POST")) {
            if(trace_level & TRACE_STD) {
                do_log(log_type, "", "process.fcgi: process POST request");
            }
            conn_length = FCGX_GetParam("CONTENT_LENGTH", cgi->envp);
            if(conn_length) {
                content_length = strtol(conn_length, NULL, 10);
                if(trace_level & TRACE_STD) {
                    do_log(log_type, "", "process.fcgi: process POST request with length: %d", content_length);
                }
                if(content_length > 0) {
                    post = (char *)malloc((content_length<<2)*sizeof(char));
                    if(post == NULL) {
                        do_log(log_type, "", "process.fcgi: unable to allocate memory for post request");
                        proc_error = NO_MEMORY_FOR_POST_REQ;
                    } else {
                        int post_index = 0, get_len = 0;
                        while((get_len = FCGX_GetStr(&post[post_index], content_length, cgi->in)) > 0) {
                            if(trace_level & TRACE_STD) {
                                do_log(log_type, "", "process.fcgi: FCGX_GetStr: %d - %d", post_index, get_len);
                            }
                            post_index += get_len;
                        }
                        if(post_index != content_length) {
                            do_log(log_type, "", "process.fcgi: POST didn't receive as many bytes as expected");
                            proc_error = INCOMPLETE_POST_REQUEST;
                        }
                        else {
                            post[content_length] = 0;
                            if(trace_level & TRACE_STD) {
                                do_log(log_type, "", "POST[%d]: %s", content_length, post);
                            }
                            query_string = post;
                        }
                    }
                    if(trace_level & TRACE_STD) {
                        do_log(log_type, "", "process.fcgi: request processed");
                    }
                } else {
                    if(trace_level & TRACE_STD) {
                        do_log(log_type, "", "process.fcgi: empty log, ignore it");
                    }
                    proc_error = ERROR_EMPTY_QUERY;
                }
            } else {
                do_log(log_type, "", "process.fcgi: content length field not present");
                proc_error = ERROR_CONTENT_LEN_FIELD_MISSING;
            }
        } else {
            do_log(log_type, "", "process.fcgi: unknown request: %s", req_method);
            proc_error = ERROR_UNKNOWN_REQUEST;
        }
    }
    if(!proc_error) {
        filename = (char *)malloc(256*sizeof(char));
        if(filename == NULL) {
            do_log(log_type, "", "process.fcgi: failed to allocate memory for filename");
            proc_error = NO_MEMORY_FOR_FILENAME;
        } else {
            if(trace_level & TRACE_STD) {
                do_log(log_type, "", "[%d] script=%s", req->ind, script_name);
            }
            if(strstr(script_name, "ap/log.fcgi")) {
                strcpy(filename, doc_root);
                strcat(filename, WEBDAV_LOG);
                proc_error = logging( filename, query_string );
            } else if(strstr(script_name, "ap/alarm.fcgi")) {
                strcpy(filename, doc_root);
                strcat(filename, WEBDAV_ALARM);
                proc_error = alarms( filename, query_string, req_method);
            } else if(strstr(script_name, "ap/cdr.fcgi")) {
                if(reporting_installed) {
                    if(strncmp(query_string, "<", 1) && strncmp(query_string, "msg=", 4)) {   // relay CDR
                        strcpy(filename, doc_root);
                        strcat(filename, WEBDAV_CDR);
                        proc_error = relay_cdr(filename, query_string);
                    } else {
                        char * msg = NULL;
                        if(!strcmp(req_method, "GET")) {    // get XML from GET query string
                            CURL * conn = curl_easy_init();
                            if(conn) {
                                msg = get_url_param(conn, query_string, "msg", "cdr", LOG_INNOVAPHONE);
                                curl_easy_cleanup(conn);
                                if(msg) 
                                    query_string = msg;
                                else 
                                    proc_error = NO_MSG_FIELD_GET_REQ;
                            }
                            else proc_error = NO_CURL_INSTANCE;
                        }
                        // syslog(LOG_INFO, "%s", query_string);
                        if(!proc_error) {
                            proc_error = cdr(req, query_string);
                        }
                        if(msg) free(msg);
                    }
                }
                else {
                    do_log(log_type, "", "process.fcgi: no access to '%s': reporting is not installed", script_name);
                    proc_error = NO_REPORTING_INSTALLED;
                }
            } else if(strstr(script_name, "apps/innovaphone-reporting/mypbx/mypbx.fcgi")) {
                if(reporting_installed) {
                    proc_error = mypbx(query_string, cgi, doc_root, &bytes_sent, &output_len);
                    external_cmd = 1;   // 
                }
                else {
                    do_log(log_type, "", "process.fcgi: no access to '%s': reporting is not installed", script_name);
                    proc_error = NO_REPORTING_INSTALLED;
                }
            } else if(strstr(script_name, "apps/innovaphone-reporting/report.fcgi") || strstr(script_name, "apps/innovaphone-reporting/user/report.fcgi")) {
                if(reporting_installed) {
                    proc_error = create_report(query_string, cgi, doc_root, &bytes_sent, &output_len);
                    external_cmd = 1;   // 
                }
                else {
                    do_log(log_type, "", "process.fcgi: no access to '%s': reporting is not installed", script_name);
                    proc_error = NO_REPORTING_INSTALLED;
                }
            } else {
                // try to find executable
                char * fcgi = strstr(script_name, ".fcgi");
                if(fcgi) {
                    char * exefile = (char *) malloc(128*sizeof(char));
                    strcpy(exefile, doc_root);
                    strncat(exefile, script_name, fcgi - script_name);
                    strcat(exefile, ".out");
                    FILE * exefp = fopen(exefile, "r");
                    if(!exefp) {
                        do_log(log_type, "", "process.fcgi: executable file %s for script %s does not exist", exefile, script_name);
                        proc_error = NO_SCRIPT_FOUND;
                    }
                    else {
                        fclose(exefp);
                        char tmp_template[] = "/tmp/query_XXXXXX";
                        int tempfile = mkstemp(tmp_template);
                        FILE * fp_tempfile = NULL;
                        if(tempfile > 0 && (fp_tempfile = fdopen(tempfile, "w+"))) {
                            fwrite(query_string, 1, strlen(query_string), fp_tempfile);
                            fclose(fp_tempfile);
                            char * cmd = (char *) malloc((strlen(exefile) + strlen(tmp_template) + 128) * sizeof(char));
                            sprintf(cmd, "%s %s", exefile, tmp_template);
                            //syslog(LOG_INFO, "calling cmd: %s ", cmd);
                            FILE * fp_cmd = popen(cmd, "r");
                            if(fp_cmd != NULL) {
                                external_cmd = 1;
                                char buffer[128];
                                while(!feof(fp_cmd)) {
                                    if(fgets(buffer, sizeof(buffer) * sizeof(char), fp_cmd) != 0) {
                                        output_len += strlen(buffer);
                                        bytes_sent += FCGX_PutS(buffer, cgi->out);
                                    }
                                }
                                pclose(fp_cmd);
                            }
                            else {
                                do_log(log_type, "", "process.fcgi: couldn't execute cmd %s", cmd);
                                proc_error = ERROR_EXECUTING_CMD;
                            }
                            free(cmd);
                            remove(tmp_template); 
                        }
                        else {
                            do_log(log_type, "", "process.fcgi: couldn't create/open tempfile");
                            proc_error = ERROR_CREATING_OPENING_FILE;
                        }
                    }
                    free(exefile);
                }
                else {
                    do_log(log_type, "", "process.fcgi: unknown script %s", script_name);
                    proc_error = NO_SCRIPT_FOUND;
                }
            }
            free(filename);
            if(trace_level & TRACE_STD) {
                do_log(log_type, "", "[%d] script=%s processed", req->ind, script_name);
            }
        }
    }
    if(post) free(post);
    if(external_cmd == 1 && output_len > 0) {
        // do nothing
    }
    else if(!proc_error) {
        bytes_sent = FCGX_PutS(STATUS_200, cgi->out);
        output_len = strlen(STATUS_200);
    } else {
        char error_len_str[11];
        int error_len;
        if(proc_error < num_errors) {
            error_len = strlen(HTTP_RESPONSE_BODY_HEAD) + strlen(errors[proc_error]) + strlen(HTTP_RESPONSE_BODY_TAIL);
        } else {
            error_len = strlen(INTERNAL_SERVER_ERROR);
        }
        sprintf(error_len_str, "%d", error_len);
        bytes_sent = FCGX_PutS(STATUS_500, cgi->out);
        bytes_sent += FCGX_PutS(CONTENT_TYPE_HTML, cgi->out);
        bytes_sent += FCGX_PutS(CONTENT_LENGTH, cgi->out);
        bytes_sent += FCGX_PutS(error_len_str, cgi->out);
        bytes_sent += FCGX_PutS("\r\n\r\n", cgi->out);
        if(proc_error < num_errors) {
            bytes_sent += FCGX_PutS(HTTP_RESPONSE_BODY_HEAD, cgi->out);
            bytes_sent += FCGX_PutS(errors[proc_error], cgi->out);
            bytes_sent += FCGX_PutS(HTTP_RESPONSE_BODY_TAIL, cgi->out);
        } else {
            bytes_sent += FCGX_PutS(INTERNAL_SERVER_ERROR, cgi->out);
        }
        output_len = strlen(STATUS_500) + strlen(CONTENT_TYPE_HTML) + strlen(CONTENT_LENGTH) + strlen(error_len_str) + 4 + error_len;
        FCGI_SetExitStatus(500);
    }
    FCGX_Finish_r(cgi);
    if(bytes_sent != output_len) {
        FCGX_ShutdownPending();
    }
    free(cgi);
    free(logfile);
    req->thread_running = 0;
    if(trace_level & TRACE_STD) {
        do_log(log_type, "", "[%d] thread finished", req->tid);
    }
    pthread_exit((void*)0);
}

void server_error(FCGX_Request *fcgi) {
    int bytes_sent = 0, output_len = 0;
	char error_len_str[11];

	sprintf(error_len_str, "%d", strlen(INTERNAL_SERVER_ERROR));
	bytes_sent = FCGX_PutS(STATUS_500, fcgi->out);
	bytes_sent += FCGX_PutS(CONTENT_TYPE_HTML, fcgi->out);
	bytes_sent += FCGX_PutS(CONTENT_LENGTH, fcgi->out);
	bytes_sent += FCGX_PutS(error_len_str, fcgi->out);
	bytes_sent += FCGX_PutS("\r\n\r\n", fcgi->out);
	bytes_sent += FCGX_PutS(INTERNAL_SERVER_ERROR, fcgi->out);
	output_len = strlen(STATUS_500) + strlen(CONTENT_TYPE_HTML) + strlen(CONTENT_LENGTH) + strlen(error_len_str) + 4 + strlen(INTERNAL_SERVER_ERROR);
	FCGI_SetExitStatus(500);
	FCGX_Finish_r (fcgi);
	if(bytes_sent != output_len) {
		FCGX_ShutdownPending();
	}
}

//clean resources
void sig_handler(int signum)
{
    do_log(LOG_REPORTING, "", "Received termination signal: %d", signum);
}

int main(int argc, char** argv) {
    int i, ind_req = 0, thread_error;
    req_t **requests;
    FCGX_Request *tmp_fcgi;
    psql_t *pg;
    pthread_mutex_t ldap_mutex;
    pthread_mutex_t cdr_queue_mutex;
    sem_t cdr_queue_count;
    struct _cdr_node **cdr_queue;
    ldap_servers_t *ldap = NULL;

    // received signal if web server is stopped 
    //signal(SIGTERM, sig_handler);

    trace_level = read_trace_level();

    if(argc > 1) {
        int bytes_sent = 0, output_len = 0, proc_error;
        if(trace_level & TRACE_STD) {
            do_log(LOG_REPORTING, "", "argc: %d %s", argc, argv[1]);
        }
        if((proc_error = create_report(argv[1], NULL, DOC_ROOT, &bytes_sent, &output_len)) != 0) {
            do_log(LOG_REPORTING, "", "Error creating report: %d", proc_error);
        }
        return 0;
    }

    if((pg = (psql_t *)malloc(sizeof(psql_t))) == NULL) {
        do_log(LOG_INNOVAPHONE, "", "error: no memory for psql");
        return 0;
    }

    pthread_mutex_init(&ldap_mutex, NULL);
    pthread_mutex_init(&cdr_queue_mutex, NULL);
    sem_init(&cdr_queue_count, 0, 0);
    cdr_queue = (struct _cdr_node **)malloc(sizeof(struct _cdr_node *));
    *cdr_queue = NULL;
 
    do_log(LOG_REPORTING, "", "create %x trace level %d", cdr_queue, trace_level);

    pg->tid = 0;
    pg->cdr_queue_mutex = &cdr_queue_mutex;
    pg->cdr_queue_count = &cdr_queue_count;
    pg->cdr_queue = cdr_queue;
    init_db_conn(&pg->conn, NULL);
    if((pg->conn == NULL) || ((ldap = read_ldap_config(pg->conn, NULL, NULL)) == NULL)) {
        create_ldap_changed_file();
    }
    if((thread_error = pthread_create(&pg->tid, 0, (void*)&postgresql_exec, pg)) != 0) {
        do_log(LOG_INNOVAPHONE, "", "error creating postgresql thread=%d", thread_error);
        return 0;
    } 
    else {
        pthread_detach(pg->tid);
    }		 

    if((requests = (req_t **)malloc(MAX_NUM_THREADS*sizeof(req_t *))) == NULL) {
        do_log(LOG_INNOVAPHONE, "", "error: no memory for process.fcgi");
        return 0;
    }
    for(i=0;i<MAX_NUM_THREADS;i++) {
        req_t *ptr_req;

        if((ptr_req = (req_t *)malloc(MAX_NUM_THREADS*sizeof(req_t))) == NULL) {
			do_log(LOG_INNOVAPHONE, "", "error: no memory for process.fcgi");
			return 0;
        }
        memset(ptr_req, 0x0, sizeof(req_t));
        ptr_req->cdr_queue = cdr_queue;
        ptr_req->ldap_mutex = &ldap_mutex;
        ptr_req->cdr_queue_mutex = &cdr_queue_mutex;
        ptr_req->cdr_queue_count = &cdr_queue_count;
        ptr_req->pg_conn = &pg->conn;
        ptr_req->ldap = &ldap;
        sem_init(&ptr_req->cdr_sem, 0, 0);

        requests[i] = ptr_req;
    }

    for (;;) {
        int run_threads = 0;
        int err = FCGX_Init();
        if (err) {
            do_log(LOG_INNOVAPHONE, "", "process.fcgi: FCGX_Init failed (error %d)", err);
            break;
        }

        if((tmp_fcgi = malloc(sizeof (FCGX_Request))) == NULL) {
            do_log(LOG_INNOVAPHONE, "", "error: no memory for tmp FCGX_Request");
            return 0;
        }
        memset(tmp_fcgi, 0x0, sizeof(FCGX_Request));

        tmp_fcgi->flags |= FCGI_FAIL_ACCEPT_ON_INTR;
        FCGX_InitRequest (tmp_fcgi, LISTENSOCK_FILENO, LISTENSOCK_FLAGS);
        if (FCGX_Accept_r (tmp_fcgi) < 0) {
            do_log(LOG_INNOVAPHONE, "", "process.fcgi: FCGX_Accept_r failed");
            break;
        }

        while(requests[ind_req]->thread_running) {
            ind_req = (ind_req == (MAX_NUM_THREADS-1) ? 0 : (ind_req + 1));
            run_threads++;
            if(run_threads == MAX_NUM_THREADS) break;
        }

        /* no threads available */
        if(run_threads == MAX_NUM_THREADS) {
            server_error(tmp_fcgi);

            free(tmp_fcgi);

            do_log(LOG_INNOVAPHONE, "", "No threads available t=%d", ind_req);
        } 
        else {
            req_t *req = requests[ind_req];

            req->fcgi_req = tmp_fcgi;
            req->ind = ind_req;
            req->thread_running = 1;

            if((thread_error = pthread_create(&req->tid, 0, (void*)&process_fcgi, req)) != 0) {
                server_error(req->fcgi_req);
                do_log(LOG_INNOVAPHONE, "", "error creating thread=%d t=%d", thread_error, ind_req);
                req->thread_running = 0;
                free(req->fcgi_req);
            } else {
                pthread_detach(req->tid);
            }		 
            ind_req = (ind_req == (MAX_NUM_THREADS-1) ? 0 : (ind_req + 1));
        }
    }

    for(i=0;i<MAX_NUM_THREADS;i++) {
        req_t *ptr_req = requests[i];

        if(ptr_req) free(ptr_req);
    }
    free(requests);

    return 0;
}
